from django.db import IntegrityError, connections, transaction
from django.db.utils import InternalError

from arkindex.documents.models import ElementPath
from arkindex.project.tests import FixtureTestCase


class TestPathConstraints(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.element_type = cls.corpus.types.first()
        cls.parent1 = cls.corpus.elements.get(name="Volume 1")
        cls.parent2 = cls.corpus.elements.get(name="Volume 2")
        cls.element = cls.corpus.elements.create(
            type=cls.element_type,
            name="Test element",
        )
        cls.child = cls.corpus.elements.create(
            type=cls.element_type,
            name="Test child",
        )

        cls.element.add_parent(cls.parent1)
        cls.element.add_parent(cls.parent2)

        cls.child.add_parent(cls.element)

    def test_unique_paths(self):
        """
        Ensure an Element cannot have strictly the same paths twice
        """
        self.assertCountEqual(self.child.paths.values_list("path", "ordering"), [
            ([self.parent1.id, self.element.id], 0),
            ([self.parent2.id, self.element.id], 0)
        ])

        # Cannot create a path that already exists
        with self.assertRaises(IntegrityError):
            self.child.paths.create(path=[self.parent1.id, self.element.id], ordering=42)

    def test_unique_element_orderings_insert(self):
        self.child.paths.get(path=[self.parent2.id, self.element.id]).delete()

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        # Creating a path should not raise anything: the trigger only runs on commit because it is deferred.
        self.child.paths.create(
            path=[self.parent2.id, self.element.id],
            # Cause the child to have two orderings within self.element
            ordering=11111,
        )

        with self.assertRaisesMessage(InternalError, "Each element may only have one ordering within the same parent"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)

    def test_unique_element_orderings_update_element(self):
        # self.child is the only child of self.element, it has two paths
        # because self.element has two parents, but it only has one ordering
        self.assertEqual(self.child.paths.count(), 2)
        self.assertCountEqual(self.child.paths.values_list("ordering", flat=True).distinct(), [0])

        path = self.child.paths.get(path=[self.parent1.id, self.element.id])

        # Assign this path to another element, and assign a different ordering
        path.element = self.corpus.elements.create(type=self.element_type, name="Another child")
        path.ordering = 65535
        path.save()

        # Nothing is wrong with the constraints yet, the path is just a little strange.
        connections["default"].check_constraints()

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        # Reassign the path to the child, without touching the ordering, so that the child now has
        # two different orderings within self.element.
        path.element = self.child
        # path.save() should not raise anything: the trigger only runs on commit because it is deferred.
        # We save using update_fields to really ensure only the element gets updated.
        path.save(update_fields=["element"])

        with self.assertRaisesMessage(InternalError, "Each element may only have one ordering within the same parent"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)

    def test_unique_element_orderings_update_path(self):
        # self.child is the only child of self.element, it has two paths
        # because self.element has two parents, but it only has one ordering
        self.assertEqual(self.child.paths.count(), 2)
        self.assertCountEqual(self.child.paths.values_list("ordering", flat=True).distinct(), [0])

        path = self.child.paths.get(path=[self.parent1.id, self.element.id])
        # Change the path in some way, so that this path no longer ends with self.element
        path.path = [self.parent1.id]
        # Change the ordering
        path.ordering = 42
        path.save()

        # Nothing is wrong with the constraints yet, the path is just a little strange.
        connections["default"].check_constraints()

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        # Restore the old path: now we are only saving to update the path, not the ordering,
        # but this will cause the child to have two orderings within self.element.
        path.path = [self.parent1.id, self.element.id]
        # path.save() should not raise anything: the trigger only runs on commit because it is deferred.
        # We save using update_fields to really ensure only the path gets updated.
        path.save(update_fields=["path"])

        with self.assertRaisesMessage(InternalError, "Each element may only have one ordering within the same parent"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)

    def test_unique_element_orderings_update_ordering(self):
        # self.child is the only child of self.element, it has two paths
        # because self.element has two parents, but it only has one ordering
        self.assertEqual(self.child.paths.count(), 2)
        self.assertCountEqual(self.child.paths.values_list("ordering", flat=True).distinct(), [0])

        path = self.child.paths.first()
        path.ordering = 42

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        # path.save() should not raise anything: the trigger only runs on commit because it is deferred.
        # We save using update_fields to really ensure only the ordering gets updated.
        path.save(update_fields=["ordering"])

        with self.assertRaisesMessage(InternalError, "Each element may only have one ordering within the same parent"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)

    def test_unique_parent_orderings_insert(self):
        self.assertEqual(self.element.paths.get(path=[self.parent2.id]).ordering, 3)

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        # Creating a path should not raise anything: the trigger only runs on commit because it is deferred.
        self.child.paths.create(
            path=[self.parent2.id],
            # Cause ordering 0 to be assigned to both self.element and self.child
            ordering=3,
        )

        with self.assertRaisesMessage(InternalError, "Each element within a parent must have a distinct ordering"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)

    def test_unique_parent_orderings_update_element(self):
        self.assertEqual(self.child.paths.count(), 2)
        self.assertCountEqual(self.child.paths.values_list("ordering", flat=True).distinct(), [0])

        path = self.child.paths.first()
        # Change the element so that one path is on self.child with ordering 0, and one path is on this element with the same ordering
        path.element = self.corpus.elements.create(type=self.element_type, name="Another child")

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        # path.save() should not raise anything: the trigger only runs on commit because it is deferred.
        # We save using update_fields to really ensure only the element gets updated.
        path.save(update_fields=["element"])

        with self.assertRaisesMessage(InternalError, "Each element within a parent must have a distinct ordering"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)

    def test_unique_parent_orderings_update_path(self):
        connections["default"].check_constraints()

        self.assertEqual(self.element.paths.get(path=[self.parent2.id]).ordering, 3)

        # Replicate the path of self.element, first with just the ordering
        # We update all of the child's paths to avoid breaking unique_element_orderings
        self.child.paths.update(ordering=3)

        # Nothing is wrong with the constraints yet, the paths are just a little strange.
        connections["default"].check_constraints()

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        path = self.child.paths.get(path=[self.parent2.id, self.element.id])
        # Now we are updating the path alone, which will cause both self.element and self.child
        # to have the same orderings within self.parent2
        path.path = [self.parent2.id]
        # path.save() should not raise anything: the trigger only runs on commit because it is deferred.
        # We save using update_fields to really ensure only the path gets updated.
        path.save(update_fields=["path"])

        with self.assertRaisesMessage(InternalError, "Each element within a parent must have a distinct ordering"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)

    def test_unique_parent_orderings_update_ordering(self):
        connections["default"].check_constraints()

        self.assertTrue(ElementPath.objects.filter(path__last=self.parent2.id, ordering=0).exists())
        self.assertEqual(self.element.paths.get(path=[self.parent2.id]).ordering, 3)

        path = self.element.paths.get(path=[self.parent2.id])
        # Update the ordering so that conflicts with the existing path
        path.ordering = 0

        # Committing the transaction should cause an error since the trigger will raise an exception.
        # Since we are within a Django TestCase, such a failure will cause the whole test class to fail,
        # so we use a savepoint to handle the failure gracefully.
        sid = transaction.savepoint()

        # path.save() should not raise anything: the trigger only runs on commit because it is deferred.
        # We save using update_fields to really ensure only the ordering gets updated.
        path.save()

        with self.assertRaisesMessage(InternalError, "Each element within a parent must have a distinct ordering"):
            # Committing the savepoint would not execute the deferred trigger, and committing the transaction
            # would mess with the test class and any subsequent unit tests, so the next best thing is to act
            # like we are about to commit by forcing all constraint checks to run.
            connections["default"].check_constraints()

        transaction.savepoint_rollback(sid)
