import logging
from textwrap import dedent

from django.conf import settings
from django.db import transaction
from django.db.models import Exists, OuterRef
from django.utils import timezone
from django.utils.module_loading import import_string
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.ponos.models import FINAL_STATES, Agent, AgentMode, Artifact, State, Task
from arkindex.ponos.signals import task_failure
from arkindex.project.serializer_fields import EnumField
from arkindex.project.triggers import notify_process_completion

logger = logging.getLogger(__name__)

TaskAgentField = import_string(getattr(settings, "TASK_AGENT_FIELD", None) or "arkindex.project.serializer_fields.NullField")
TaskGPUField = import_string(getattr(settings, "TASK_GPU_FIELD", None) or "arkindex.project.serializer_fields.NullField")


class TaskLightSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance without logs or agent information.
    Used to list tasks inside a process.
    """
    state = EnumField(
        State,
        help_text=dedent("""
            Allowed transitions for the state of a task by an agent are defined below:

                Pending ⟶ Running ⟶ Completed
                  └⟶ Error   ├⟶ Failed
                             └⟶ Error
                Stopping ⟶ Stopped
                  └⟶ Error
            Slurm agents are also allowed to update state from Pending to Completed or Failed.
        """).strip(),
    )

    class Meta:
        model = Task
        fields = (
            "id",
            "run",
            "depth",
            "slug",
            "state",
            "parents",
            "shm_size",
            "requires_gpu",
        )
        read_only_fields = (
            "id",
            "run",
            "depth",
            "slug",
            "parents",
            "shm_size",
            "requires_gpu",
        )

    def validate_state(self, state):
        # Updates from a state to the same state is blocked to avoid side effects on finished tasks
        allowed_transitions = {
            State.Unscheduled: [State.Pending],
            State.Pending: [State.Running, State.Error],
            State.Running: [State.Completed, State.Failed, State.Error],
            State.Stopping: [State.Stopped, State.Error],
        }
        user = self.context["request"].user
        if isinstance(user, Agent) and user.mode == AgentMode.Slurm:
            allowed_transitions[State.Pending].extend([State.Completed, State.Failed])
        if self.instance and state not in allowed_transitions.get(self.instance.state, []):
            raise ValidationError(f"Transition from state {self.instance.state} to state {state} is forbidden.")
        return state


class TaskSerializer(TaskLightSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance with logs and agent information.
    """

    logs = serializers.SerializerMethodField()
    full_log = serializers.SerializerMethodField()
    agent = TaskAgentField(read_only=True, allow_null=True)
    gpu = TaskGPUField(read_only=True, allow_null=True)
    extra_files = serializers.DictField(read_only=True)

    class Meta(TaskLightSerializer.Meta):
        fields = TaskLightSerializer.Meta.fields + (
            "logs",
            "full_log",
            "agent",
            "gpu",
            "extra_files",
            "original_task_id"
        )
        read_only_fields = TaskLightSerializer.Meta.read_only_fields + (
            "logs",
            "full_log",
            "agent",
            "gpu",
            "extra_files",
            "original_task_id"
        )

    @extend_schema_field(serializers.CharField())
    def get_logs(self, task) -> str:
        return task.logs.latest

    @extend_schema_field(serializers.URLField())
    def get_full_log(self, task) -> str:
        return task.logs.s3_url

    def update(self, instance: Task, validated_data) -> Task:
        """
        Perform updates on a task instance.
        If a task gets marked as :attr:`~arkindex.ponos.models.State.Completed`, this will set
        its child tasks to :attr:`~arkindex.ponos.models.State.Pending` if all of their dependencies are met.
        """
        instance = super().update(instance, validated_data)

        if instance.state not in FINAL_STATES:
            return instance

        with transaction.atomic(using="default"):
            # When a task is completed, set its children to be started next
            if instance.state == State.Completed:
                (
                    instance.children
                    # We need to use the default Database here to avoid a stale read
                    # as the parent instance has been updated just before
                    .using("default")
                    .filter(state=State.Unscheduled)
                    # Do not update any child whose dependencies include some tasks that are not completed
                    .exclude(Exists(Task.objects.filter(children=OuterRef("pk")).exclude(state=State.Completed)))
                    .update(state=State.Pending)
                )

            # When some tasks have failed, tasks that depend on them will remain unscheduled.
            # When a task is completed, its children will be set to pending.
            # All children have a depth above zero, since they have at least one dependency.
            # This means a process can be considered finished when no task is pending (about to run) or running,
            # except for tasks with a depth of zero, which can be unscheduled until they get assigned to an agent.
            if not (
                instance.process.tasks
                .using("default")
                .filter(run=instance.run)
                .exclude(state__in=FINAL_STATES)
                .exclude(state=State.Unscheduled, depth__gt=0)
                .exists()
            ):
                instance.process.finished = timezone.now()
                instance.process.save()

                notify_process_completion(instance.process)

        # We already checked earlier that the task was in a final state.
        # If this state is both final and not completed, then we should trigger the task failure signal.
        if instance.state != State.Completed:
            task_failure.send_robust(self.__class__, task=instance)

        return instance


class TaskTinySerializer(TaskSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance with only its state.
    Used by humans to update a task.
    """

    state = EnumField(
        State,
        help_text="The state can only be updated from `running` to `stopping`, to manually stop a task.",
    )

    class Meta:
        model = Task
        fields = ("id", "state")
        read_only_fields = ("id",)

    def validate_state(self, state):
        """
        Only allow a user to manually stop a task
        """
        if (
            self.instance
            and self.instance.state != state
            and (self.instance.state != State.Running or state != State.Stopping)
        ):
            raise ValidationError("State can only be updated from running to stopping")
        return state

    def update(self, instance: Task, validated_data) -> Task:
        new_state = validated_data.get("state")

        if new_state in FINAL_STATES and new_state != State.Completed:
            task_failure.send_robust(self.__class__, task=instance)

        return super().update(instance, validated_data)


class ArtifactSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Artifact` instance to allow
    agents to create them, and users to list and retrieve them
    """

    s3_put_url = serializers.SerializerMethodField()

    class Meta:
        model = Artifact
        fields = (
            "id",
            "path",
            "size",
            "content_type",
            "s3_put_url",
            "created",
            "updated",
        )
        read_only_fields = (
            "id",
            "s3_put_url",
            "created",
            "updated",
        )

    @extend_schema_field(serializers.URLField(allow_null=True))
    def get_s3_put_url(self, obj):
        """Only add the PUT url to store the file for agents during creation"""
        request = self.context.get("request")
        if not request or request.method != "POST":
            return
        return obj.s3_put_url

    def validate_path(self, path):
        """Check that no artifacts with this path already exist in DB"""
        task = self.context["view"].task
        if task.artifacts.filter(path=path).exists():
            raise ValidationError("An artifact with this path already exists")

        return path
